import unittest

from pony.orm import *
from pony.orm.tests.testutils import *

db = Database('sqlite', ':memory:')

sep = ' '

class Person(db.Entity):
    id = PrimaryKey(int)
    first_name = Required(str)
    last_name = Required(str)
    favorite_color = Optional(str)
    cars = Set(lambda: Car)

    @property
    def full_name(self):
        return self.first_name + sep + self.last_name

    @property
    def full_name_2(self):
        return concat(self.first_name, sep, self.last_name)  # tests using of function `concat` from external scope

    @property
    def has_car(self):
        return not self.cars.is_empty()

    def cars_by_color1(self, color):
        return select(car for car in self.cars if car.color == color)

    def cars_by_color2(self, color):
        return self.cars.select(lambda car: car.color == color)

    @property
    def cars_price(self):
        return sum(c.price for c in self.cars)

    @property
    def incorrect_full_name(self):
        return self.first_name + ' ' + p.last_name  # p is FakePerson instance here

    @classmethod
    def find_by_full_name(cls, full_name):
        return cls.select(lambda p: p.full_name_2 == full_name)


class FakePerson(object):
    pass

p = FakePerson()
p.last_name = '***'


class Car(db.Entity):
    brand = Required(str)
    model = Required(str)
    owner = Optional(Person)
    year = Required(int)
    price = Required(int)
    color = Required(str)

db.generate_mapping(create_tables=True)

with db_session:
    p1 = Person(id=1, first_name='Alexander', last_name='Kozlovsky', favorite_color='white')
    p2 = Person(id=2, first_name='Alexei', last_name='Malashkevich', favorite_color='green')
    p3 = Person(id=3, first_name='Vitaliy', last_name='Abetkin')
    p4 = Person(id=4, first_name='Alexander', last_name='Tischenko', favorite_color='blue')

    c1 = Car(brand='Peugeot', model='306', owner=p1, year=2006, price=14000, color='red')
    c2 = Car(brand='Honda', model='Accord', owner=p1, year=2007, price=13850, color='white')
    c3 = Car(brand='Nissan', model='Skyline', owner=p2, year=2008, price=29900, color='black')
    c4 = Car(brand='Volkswagen', model='Passat', owner=p1, year=2012, price=9400, color='blue')
    c5 = Car(brand='Koenigsegg', model='CCXR', owner=p4, year=2016, price=4850000, color='white')
    c6 = Car(brand='Lada', model='Kalina', owner=p4, year=2015, price=5000, color='white')


class TestHybridsAndProperties(unittest.TestCase):
    @db_session
    def test1(self):
        persons = select(p.full_name for p in Person if p.has_car)[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky', 'Alexei Malashkevich', 'Alexander Tischenko'})

    @db_session
    def test2(self):
        cars_prices = select(p.cars_price for p in Person)[:]
        self.assertEqual(set(cars_prices), {0, 29900, 37250, 4855000})

    @db_session
    def test3(self):
        persons = select(p.full_name for p in Person if p.cars_price > 100000)[:]
        self.assertEqual(set(persons), {'Alexander Tischenko'})

    @db_session
    def test4(self):
        persons = select(p.full_name for p in Person if not p.cars_price)[:]
        self.assertEqual(set(persons), {'Vitaliy Abetkin'})

    @db_session
    def test5(self):
        persons = select(p.full_name for p in Person if exists(c for c in p.cars_by_color2('white') if c.price > 10000))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky', 'Alexander Tischenko'})

    @db_session
    def test6(self):
        persons = select(p.full_name for p in Person if exists(c for c in p.cars_by_color1('white') if c.price > 10000))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky', 'Alexander Tischenko'})

    @db_session
    def test7(self):
        c1 = Car[1]
        persons = select(p.full_name for p in Person if c1 in p.cars_by_color2('red'))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky'})

    @db_session
    def test8(self):
        c1 = Car[1]
        persons = select(p.full_name for p in Person if c1 in p.cars_by_color1('red'))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky'})

    @db_session
    def test9(self):
        persons = select(p.full_name for p in Person if p.cars_by_color1(p.favorite_color))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky'})

    @db_session
    def test10(self):
        persons = select(p.full_name for p in Person if not p.cars_by_color1(p.favorite_color))[:]
        self.assertEqual(set(persons), {'Alexander Tischenko', 'Alexei Malashkevich', 'Vitaliy Abetkin'})

    @db_session
    def test11(self):
        persons = select(p.full_name for p in Person if p.cars_by_color2(p.favorite_color))[:]
        self.assertEqual(set(persons), {'Alexander Kozlovsky'})

    @db_session
    def test12(self):
        persons = select(p.full_name for p in Person if not p.cars_by_color2(p.favorite_color))[:]
        self.assertEqual(set(persons), {'Alexander Tischenko', 'Alexei Malashkevich', 'Vitaliy Abetkin'})

    @db_session
    def test13(self):
        persons = select(p.full_name for p in Person if count(p.cars_by_color1('white')) > 1)
        self.assertEqual(set(persons), {'Alexander Tischenko'})

    @db_session
    def test14(self):
        # This test checks if accessing function-specific globals works correctly
        persons = select(p.incorrect_full_name for p in Person if p.has_car)[:]
        self.assertEqual(set(persons), {'Alexander ***', 'Alexei ***', 'Alexander ***'})

    @db_session
    def test15(self):
        # Test repeated use of the same generator with hybrid method/property that uses funciton from external scope
        result = Person.find_by_full_name('Alexander Kozlovsky')
        self.assertEqual(set(obj.last_name for obj in result), {'Kozlovsky'})
        result = Person.find_by_full_name('Alexander Kozlovsky')
        self.assertEqual(set(obj.last_name for obj in result), {'Kozlovsky'})
        result = Person.find_by_full_name('Alexander Tischenko')
        self.assertEqual(set(obj.last_name for obj in result), {'Tischenko'})

    @db_session
    def test16(self):
        result = Person.select(lambda p: p.full_name == 'Alexander Kozlovsky')
        self.assertEqual(set(p.id for p in result), {1})

    @db_session
    def test17(self):
        global sep
        sep = '.'
        try:
            result = Person.select(lambda p: p.full_name == 'Alexander.Kozlovsky')
            self.assertEqual(set(p.id for p in result), {1})
        finally:
            sep = ' '

    @db_session
    def test18(self):
        result = Person.select().filter(lambda p: p.full_name == 'Alexander Kozlovsky')
        self.assertEqual(set(p.id for p in result), {1})

    @db_session
    def test19(self):
        global sep
        sep = '.'
        try:
            result = Person.select().filter(lambda p: p.full_name == 'Alexander.Kozlovsky')
            self.assertEqual(set(p.id for p in result), {1})
        finally:
            sep = ' '

if __name__ == '__main__':
    unittest.main()





import unittest


from pony.orm import *

db = Database()

class TestPost(db.Entity):
    category = Optional('TestCategory')
    name = Optional(str, default='Noname')

class TestCategory(db.Entity):
    posts = Set(TestPost)

db.bind('sqlite', ':memory:')
db.generate_mapping(create_tables=True)


class EntityStatusTestCase(object):

    def make_flush(self, obj=None):
        raise NotImplementedError

    @db_session
    def test_delete_updated(self):
        p = TestPost()
        self.make_flush(p)
        p.name = 'Pony'
        self.assertEqual(p._status_, 'modified')
        self.make_flush(p)
        self.assertEqual(p._status_, 'updated')
        p.delete()
        self.assertEqual(p._status_, 'marked_to_delete')
        self.make_flush(p)
        self.assertEqual(p._status_, 'deleted')

    @db_session
    def test_delete_inserted(self):
        p = TestPost()
        self.assertEqual(p._status_, 'created')
        self.make_flush(p)
        self.assertEqual(p._status_, 'inserted')
        p.delete()

    @db_session
    def test_cancelled(self):
        p = TestPost()
        self.assertEqual(p._status_, 'created')
        p.delete()
        self.assertEqual(p._status_, 'cancelled')
        self.make_flush(p)
        self.assertEqual(p._status_, 'cancelled')


class EntityStatusTestCase_ObjectFlush(EntityStatusTestCase,
                                       unittest.TestCase):

    def make_flush(self, obj=None):
        obj.flush()


class EntityStatusTestCase_FullFlush(EntityStatusTestCase,
                                     unittest.TestCase):

    def make_flush(self, obj=None):
        flush() # full flush



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import *

class TestOneToManyRequired(unittest.TestCase):

    def setUp(self):
        db = Database('sqlite', ':memory:', create_db=True)

        class Student(db.Entity):
            id = PrimaryKey(int)
            name = Required(unicode)
            group = Required('Group')

        class Group(db.Entity):
            number = PrimaryKey(int)
            students = Set(Student)

        self.db = db
        self.Group = Group
        self.Student = Student

        db.generate_mapping(create_tables=True)

        with db_session:
            g101 = Group(number=101)
            g102 = Group(number=102)
            g103 = Group(number=103)
            s1 = Student(id=1, name='Student1', group=g101)
            s2 = Student(id=2, name='Student2', group=g101)
            s3 = Student(id=3, name='Student3', group=g102)
            s4 = Student(id=4, name='Student3', group=g102)

        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    @raises_exception(ValueError, 'Attribute Student[1].group is required')
    def test_1(self):
        self.Student[1].group = None

    def test_2(self):
        Student, Group = self.Student, self.Group
        s1 = Student[1]
        g = Group[101]
        g.students.remove(s1)
        self.assertEqual(s1._status_, 'marked_to_delete')

    def test_3(self):
        Student, Group = self.Student, self.Group
        s1, s2 = Student[1], Student[2]
        g = Group[101]
        g.students.clear()
        self.assertEqual(s1._status_, 'marked_to_delete')
        self.assertEqual(s2._status_, 'marked_to_delete')
        self.assertEqual(set(g.students), set())

    def test_4(self):
        Student, Group = self.Student, self.Group
        s1, s2, s3, s4 = Student.select().order_by(Student.id)
        g1, g2 = Group[101], Group[102]
        g1.students = g2.students
        self.assertEqual(set(g1.students), {s3, s4})
        self.assertEqual(s1._status_, 'marked_to_delete')
        self.assertEqual(s2._status_, 'marked_to_delete')

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_5(self):
        Group, Student = self.Group, self.Student
        g = Group[101]
        g.students.add(None)

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_6(self):
        Group, Student = self.Group, self.Student
        g = Group[101]
        g.students.remove(None)

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_7(self):
        Group = self.Group
        g104 = Group(number=104, students=None)

    def test_8(self):
        db, Group, Student = self.db, self.Group, self.Student

        g = Group[101]
        s3 = Student[3]  # s3 is loaded now
        db._dblocal.last_sql = None
        g.students.add(s3)
        # Group.students.load should not attempt to load s3 from db
        self.assertEqual(db.last_sql, None)

    def test_9(self):
        db, Group, Student = self.db, self.Group, self.Student

        g = Group[101]
        e = g.students.is_empty()
        self.assertEqual(e, False)

        db._dblocal.last_sql = None
        e = g.students.is_empty()  # should take result from the cache
        self.assertEqual(e, False)
        self.assertEqual(db.last_sql, None)

        g = Group[103]
        e = g.students.is_empty()  # should take SQL from the SQL cache
        self.assertEqual(e, True)

        db._dblocal.last_sql = None
        e = g.students.is_empty()  # should take result from the cache
        self.assertEqual(e, True)
        self.assertEqual(db.last_sql, None)

    def test_10(self):
        db, Group = self.db, self.Group

        g = Group[101]
        c = len(g.students)
        self.assertEqual(c, 2)
        db._dblocal.last_sql = None
        e = g.students.is_empty()  # should take result from the cache
        self.assertEqual(e, False)
        self.assertEqual(db.last_sql, None)

        g = Group[102]
        c = g.students.count()
        self.assertEqual(c, 2)
        db._dblocal.last_sql = None
        e = g.students.is_empty()  # should take result from the cache
        self.assertEqual(e, False)
        self.assertEqual(db.last_sql, None)

        g = Group[103]
        c = len(g.students)
        self.assertEqual(c, 0)
        db._dblocal.last_sql = None
        e = g.students.is_empty()  # should take result from the cache
        self.assertEqual(e, True)
        self.assertEqual(db.last_sql, None)

    def test_11(self):
        db, Group, Student = self.db, self.Group, self.Student

        g = Group[101]
        s3 = Student[3]
        c = g.students.count()
        self.assertEqual(c, 2)

        db._dblocal.last_sql = None
        c = g.students.count()  # should take count from the cache
        self.assertEqual(c, 2)
        self.assertEqual(db.last_sql, None)

        g.students.add(s3)
        c = g.students.count()  # should take modified count from the cache
        self.assertEqual(c, 3)
        self.assertEqual(db.last_sql, None)

        g2 = Group[102]
        c = g2.students.count()  # should send query to the database
        self.assertEqual(c, 1)
        self.assertTrue(db.last_sql is not None)

    def test_12(self):
        Group, Student = self.Group, self.Student
        g = Group[101]

        s1 = Student[1]
        self.assertEqual(s1._rbits_, 0)
        self.assertTrue(s1 in g.students)
        self.assertEqual(s1._rbits_, Student._bits_[Student.group])

        s3 = Student[3]
        self.assertEqual(s3._rbits_, 0)
        self.assertTrue(s3 not in g.students)
        self.assertEqual(s3._rbits_, Student._bits_[Student.group])

        s5 = Student(id=5, name='Student5', group=g)
        self.assertEqual(s5._rbits_, None)
        self.assertTrue(s5 in g.students)
        self.assertEqual(s5._rbits_, None)

class TestOneToManyOptional(unittest.TestCase):

    def setUp(self):
        db = Database('sqlite', ':memory:', create_db=True)

        class Student(db.Entity):
            id = PrimaryKey(int)
            name = Required(unicode)
            group = Optional('Group')

        class Group(db.Entity):
            number = PrimaryKey(int)
            students = Set(Student)

        self.db = db
        self.Group = Group
        self.Student = Student

        db.generate_mapping(create_tables=True)

        with db_session:
            g101 = Group(number=101)
            g102 = Group(number=102)
            g103 = Group(number=103)
            s1 = Student(id=1, name='Student1', group=g101)
            s2 = Student(id=2, name='Student2', group=g101)
            s3 = Student(id=3, name='Student3', group=g102)
            s4 = Student(id=4, name='Student3', group=g102)

        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_1(self):
        self.Student[1].group = None
        self.assertEqual(set(self.Group[101].students), {self.Student[2]})

    def test_2(self):
        Student, Group = self.Student, self.Group
        s1 = Student[1]
        g = Group[101]
        g.students.remove(s1)
        self.assertEqual(s1.group, None)

    def test_3(self):
        Student, Group = self.Student, self.Group
        s1, s2 = Student[1], Student[2]
        g = Group[101]
        g.students.clear()
        self.assertEqual(s1.group, None)
        self.assertEqual(s2.group, None)
        self.assertEqual(set(g.students), set())

    def test_4(self):
        Student, Group = self.Student, self.Group
        s1, s2, s3, s4 = Student.select().order_by(Student.id)
        g1, g2 = Group[101], Group[102]
        g1.students = g2.students
        self.assertEqual(set(g1.students), {s3, s4})
        self.assertEqual(s1.group, None)
        self.assertEqual(s2.group, None)

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_5(self):
        Group, Student = self.Group, self.Student
        g = Group[101]
        g.students.add(None)

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_6(self):
        Group, Student = self.Group, self.Student
        g = Group[101]
        g.students.remove(None)

    @raises_exception(ValueError, 'A single Student instance or Student iterable is expected. Got: None')
    def test_7(self):
        Group = self.Group
        g104 = Group(number=104, students=None)


if __name__ == '__main__':
    unittest.main()



import unittest

from pony import orm
from pony.py23compat import buffer

db = orm.Database('sqlite', ':memory:')

class Foo(db.Entity):
    id = orm.PrimaryKey(int)
    b = orm.Optional(orm.buffer)

class Bar(db.Entity):
    b = orm.PrimaryKey(orm.buffer)

class Baz(db.Entity):
    id = orm.PrimaryKey(int)
    b = orm.Optional(orm.buffer, unique=True)

db.generate_mapping(create_tables=True)

buf = orm.buffer(b'123')

with orm.db_session:
    Foo(id=1, b=buf)
    Bar(b=buf)
    Baz(id=1, b=buf)


class Test(unittest.TestCase):
    def test_1(self):  # Bug #355
        with orm.db_session:
            Bar[buf]

    def test_2(self):  # Regression after #355 fix
        with orm.db_session:
            result = orm.select(bar.b for bar in Foo)[:]
            self.assertEqual(result, [buf])

    def test_3(self):  # Bug #390
        with orm.db_session:
            Baz.get(b=buf)



import unittest

from pony.orm import *
from pony import orm

class TestJoin(unittest.TestCase):

    exclude_fixtures = {'test': ['clear_tables']}

    @classmethod
    def setUpClass(cls):
        db = cls.db = Database('sqlite', ':memory:')

        class Genre(db.Entity):
            name = orm.Optional(str) # TODO primary key
            artists = orm.Set('Artist')
            favorite = orm.Optional(bool)
            index = orm.Optional(int)

        class Hobby(db.Entity):
            name = orm.Required(str)
            artists = orm.Set('Artist')

        class Artist(db.Entity):
            name = orm.Required(str)
            age = orm.Optional(int)
            hobbies = orm.Set(Hobby)
            genres = orm.Set(Genre)

        db.generate_mapping(create_tables=True)

        with orm.db_session:
            pop = Genre(name='pop')
            rock = Genre(name='rock')
            Artist(name='Sia', age=40, genres=[pop, rock])
            Artist(name='Lady GaGa', age=30, genres=[pop])

        pony.options.INNER_JOIN_SYNTAX = True

    @db_session
    def test_join_1(self):
        result = select(g.id for g in self.db.Genre for a in g.artists if a.name.startswith('S'))[:]
        self.assertEqual(self.db.last_sql, """SELECT DISTINCT "g"."id"
FROM "Genre" "g"
  INNER JOIN "Artist_Genre" "t-1"
    ON "g"."id" = "t-1"."genre"
  INNER JOIN "Artist" "a"
    ON "t-1"."artist" = "a"."id"
WHERE "a"."name" LIKE 'S%'""")

    @db_session
    def test_join_2(self):
        result = select(g.id for g in self.db.Genre for a in self.db.Artist
                        if JOIN(a in g.artists) and a.name.startswith('S'))[:]
        self.assertEqual(self.db.last_sql, """SELECT DISTINCT "g"."id"
FROM "Genre" "g"
  INNER JOIN "Artist_Genre" "t-1"
    ON "g"."id" = "t-1"."genre", "Artist" "a"
WHERE "t-1"."artist" = "a"."id"
  AND "a"."name" LIKE 'S%'""")


    @db_session
    def test_join_3(self):
        result = select(g.id for g in self.db.Genre for x in self.db.Artist for a in self.db.Artist
                        if JOIN(a in g.artists) and a.name.startswith('S') and g.id == x.id)[:]
        self.assertEqual(self.db.last_sql, '''SELECT DISTINCT "g"."id"
FROM "Genre" "g"
  INNER JOIN "Artist_Genre" "t-1"
    ON "g"."id" = "t-1"."genre", "Artist" "x", "Artist" "a"
WHERE "t-1"."artist" = "a"."id"
  AND "a"."name" LIKE 'S%'
  AND "g"."id" = "x"."id"''')

    @db_session
    def test_join_4(self):
        result = select(g.id for g in self.db.Genre for a in self.db.Artist for x in self.db.Artist
                        if JOIN(a in g.artists) and a.name.startswith('S') and g.id == x.id)[:]
        self.assertEqual(self.db.last_sql, '''SELECT DISTINCT "g"."id"
FROM "Genre" "g"
  INNER JOIN "Artist_Genre" "t-1"
    ON "g"."id" = "t-1"."genre", "Artist" "a", "Artist" "x"
WHERE "t-1"."artist" = "a"."id"
  AND "a"."name" LIKE 'S%'
  AND "g"."id" = "x"."id"''')

if __name__ == '__main__':
    unittest.main()



from pony.py23compat import StringIO

import unittest

from pony import orm


db = orm.Database('sqlite', ':memory:')

class A(db.Entity):
    id = orm.PrimaryKey(int)
    x = orm.Required(bool)
    y = orm.Required(float)

db.generate_mapping(create_tables=True)

with orm.db_session:
    a1 = A(id=1, x=False, y=3.0)
    a2 = A(id=2, x=True, y=4.0)
    a3 = A(id=3, x=False, y=1.0)


class TestDeduplication(unittest.TestCase):
    @orm.db_session
    def test_1(self):
        a2 = A.get(id=2)
        a1 = A.get(id=1)
        self.assertIs(a1.id, 1)

    @orm.db_session
    def test_2(self):
        a3 = A.get(id=3)
        a1 = A.get(id=1)
        self.assertIs(a1.id, 1)

    @orm.db_session
    def test_3(self):
        q = A.select().order_by(-1)
        stream = StringIO()
        q.show(stream=stream)
        s = stream.getvalue()
        self.assertEqual(s, 'id|x    |y  \n'  
                            '--+-----+---\n'
                            '3 |False|1.0\n'
                            '2 |True |4.0\n'
                            '1 |False|3.0\n')






import unittest

from pony import orm

class Test(unittest.TestCase):
    def test_1(self):
        db = orm.Database('sqlite', ':memory:')

        class Person(db.Entity):
            name = orm.Required(str)
            group = orm.Optional(lambda: Group)

        class Group(db.Entity):
            title = orm.PrimaryKey(str)
            persons = orm.Set(Person)

            def __len__(self):
                return len(self.persons)

        db.generate_mapping(create_tables=True)

        with orm.db_session:
            p1 = Person(name="Alex")
            p2 = Person(name="Brad")
            p3 = Person(name="Chad")
            p4 = Person(name="Dylan")
            p5 = Person(name="Ethan")

            g1 = Group(title="Foxes")
            g2 = Group(title="Gorillas")

            g1.persons.add(p1)
            g1.persons.add(p2)
            g1.persons.add(p3)
            g2.persons.add(p4)
            g2.persons.add(p5)
            orm.commit()

            foxes = Group['Foxes']
            gorillas = Group['Gorillas']

            self.assertEqual(len(foxes), 3)
            self.assertEqual(len(gorillas), 2)



import sys, unittest
from decimal import Decimal
from datetime import date

from pony.orm import *
from pony.orm.tests.testutils import *

class TestIndexes(unittest.TestCase):
    def test_1(self):
        db = Database('sqlite', ':memory:')
        class Person(db.Entity):
            name = Required(str)
            age = Required(int)
            composite_key(name, 'age')
        db.generate_mapping(create_tables=True)

        i1, i2 = Person._indexes_
        self.assertEqual(i1.attrs, (Person.id,))
        self.assertEqual(i1.is_pk, True)
        self.assertEqual(i1.is_unique, True)
        self.assertEqual(i2.attrs, (Person.name, Person.age))
        self.assertEqual(i2.is_pk, False)
        self.assertEqual(i2.is_unique, True)

        table = db.schema.tables['Person']
        name_column = table.column_dict['name']
        age_column = table.column_dict['age']
        self.assertEqual(len(table.indexes), 2)
        db_index = table.indexes[name_column, age_column]
        self.assertEqual(db_index.is_pk, False)
        self.assertEqual(db_index.is_unique, True)

    def test_2(self):
        db = Database('sqlite', ':memory:')
        class Person(db.Entity):
            name = Required(str)
            age = Required(int)
            composite_index(name, 'age')
        db.generate_mapping(create_tables=True)

        i1, i2 = Person._indexes_
        self.assertEqual(i1.attrs, (Person.id,))
        self.assertEqual(i1.is_pk, True)
        self.assertEqual(i1.is_unique, True)
        self.assertEqual(i2.attrs, (Person.name, Person.age))
        self.assertEqual(i2.is_pk, False)
        self.assertEqual(i2.is_unique, False)

        table = db.schema.tables['Person']
        name_column = table.column_dict['name']
        age_column = table.column_dict['age']
        self.assertEqual(len(table.indexes), 2)
        db_index = table.indexes[name_column, age_column]
        self.assertEqual(db_index.is_pk, False)
        self.assertEqual(db_index.is_unique, False)

        create_script = db.schema.generate_create_script()
        index_sql = 'CREATE INDEX "idx_person__name_age" ON "Person" ("name", "age")'
        self.assertTrue(index_sql in create_script)

    def test_3(self):
        db = Database('sqlite', ':memory:')
        class User(db.Entity):
            name = Required(str, unique=True)

        db.generate_mapping(create_tables=True)

        with db_session:
            u = User(id=1, name='A')

        with db_session:
            u = User[1]
            u.name = 'B'

        with db_session:
            u = User[1]
            self.assertEqual(u.name, 'B')

    def test_4(self):  # issue 321
        db = Database('sqlite', ':memory:')
        class Person(db.Entity):
            name = Required(str)
            age = Required(int)
            composite_key(name, age)

        db.generate_mapping(create_tables=True)
        with db_session:
            p1 = Person(id=1, name='John', age=19)

        with db_session:
            p1 = Person[1]
            p1.set(name='John', age=19)
            p1.delete()

    def test_5(self):
        db = Database('sqlite', ':memory:')

        class Table1(db.Entity):
            name = Required(str)
            table2s = Set('Table2')

        class Table2(db.Entity):
            height = Required(int)
            length = Required(int)
            table1 = Optional('Table1')
            composite_key(height, length, table1)

        db.generate_mapping(create_tables=True)

        with db_session:
            Table2(height=2, length=1)
            Table2.exists(height=2, length=1)

if __name__ == '__main__':
    unittest.main()



import unittest

from pony import orm

class Test(unittest.TestCase):
    def test_1(self):
        db = orm.Database('sqlite', ':memory:')

        class Person(db.Entity):
            id = orm.PrimaryKey(int, auto=True)
            name = orm.Required(str)
            orm.composite_key(id, name)

        db.generate_mapping(create_tables=True)

        table = db.schema.tables[Person._table_]
        pk_column = table.column_dict[Person.id.column]
        self.assertTrue(pk_column.is_pk)

        with orm.db_session:
            p1 = Person(name='John')
            p2 = Person(name='Mike')



from pony.orm import Database, Required, Optional
from flask_login import UserMixin
from datetime import datetime

db = Database()

class User(db.Entity, UserMixin):
    login = Required(str, unique=True)
    password = Required(str)
    last_login = Optional(datetime)


from __future__ import absolute_import, print_function, division
from pony.py23compat import PY2, PYPY, PYPY2

import sys, unittest
from datetime import date, datetime
from decimal import Decimal

from pony.orm.core import *
from pony.orm.sqltranslation import IncomparableTypesError
from pony.orm.tests.testutils import *

db = Database('sqlite', ':memory:')

class Student(db.Entity):
    id = PrimaryKey(int)
    name = Required(unicode)
    dob = Required(date)
    last_visit = Required(datetime)
    scholarship = Required(Decimal, 6, 2)
    phd = Required(bool)
    group = Required('Group')

class Group(db.Entity):
    number = PrimaryKey(int)
    students = Set(Student)


db.generate_mapping(create_tables=True)

with db_session:
    g1 = Group(number=1)
    g2 = Group(number=2)

    Student(id=1, name="AA", dob=date(1981, 1, 1), last_visit=datetime(2011, 1, 1, 11, 11, 11),
                   scholarship=Decimal("0"), phd=True, group=g1)

    Student(id=2, name="BB", dob=date(1982, 2, 2), last_visit=datetime(2011, 2, 2, 12, 12, 12),
                   scholarship=Decimal("202.2"), phd=True, group=g1)

    Student(id=3, name="CC", dob=date(1983, 3, 3), last_visit=datetime(2011, 3, 3, 13, 13, 13),
                   scholarship=Decimal("303.3"), phd=False, group=g1)

    Student(id=4, name="DD", dob=date(1984, 4, 4), last_visit=datetime(2011, 4, 4, 14, 14, 14),
                   scholarship=Decimal("404.4"), phd=False, group=g2)

    Student(id=5, name="EE", dob=date(1985, 5, 5), last_visit=datetime(2011, 5, 5, 15, 15, 15),
                   scholarship=Decimal("505.5"), phd=False, group=g2)


class TestFuncMonad(unittest.TestCase):
    def setUp(self):
        rollback()
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()
    def test_minmax1(self):
        result = set(select(s for s in Student if max(s.id, 3) == 3 ))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_minmax2(self):
        result = set(select(s for s in Student if min(s.id, 3) == 3 ))
        self.assertEqual(result, {Student[4], Student[5], Student[3]})
    def test_minmax3(self):
        result = set(select(s for s in Student if max(s.name, "CC") == "CC" ))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_minmax4(self):
        result = set(select(s for s in Student if min(s.name, "CC") == "CC" ))
        self.assertEqual(result, {Student[4], Student[5], Student[3]})
    def test_minmax5(self):
        x = chr(128)
        try: result = set(select(s for s in Student if min(s.name, x) == "CC" ))
        except TypeError as e:
            self.assertTrue(PY2 and e.args[0] == "The bytestring '\\x80' contains non-ascii symbols. Try to pass unicode string instead")
        else: self.assertFalse(PY2)
    def test_minmax6(self):
        x = chr(128)
        try: result = set(select(s for s in Student if min(s.name, x, "CC") == "CC" ))
        except TypeError as e:
            self.assertTrue(PY2 and e.args[0] == "The bytestring '\\x80' contains non-ascii symbols. Try to pass unicode string instead")
        else: self.assertFalse(PY2)
    def test_minmax7(self):
        result = set(select(s for s in Student if min(s.phd, 2) == 2 ))
    def test_date_func1(self):
        result = set(select(s for s in Student if s.dob >= date(1983, 3, 3)))
        self.assertEqual(result, {Student[3], Student[4], Student[5]})
    # @raises_exception(ExprEvalError, "date(1983, 'three', 3) raises TypeError: an integer is required")
    @raises_exception(TypeError, "'month' argument of date(year, month, day) function must be of 'int' type. "
                                 "Got: '%s'" % unicode.__name__)
    def test_date_func2(self):
        result = set(select(s for s in Student if s.dob >= date(1983, 'three', 3)))
    # @raises_exception(NotImplementedError)
    # def test_date_func3(self):
    #     d = 3
    #     result = set(select(s for s in Student if s.dob >= date(1983, d, 3)))
    def test_datetime_func1(self):
        result = set(select(s for s in Student if s.last_visit >= date(2011, 3, 3)))
        self.assertEqual(result, {Student[3], Student[4], Student[5]})
    def test_datetime_func2(self):
        result = set(select(s for s in Student if s.last_visit >= datetime(2011, 3, 3)))
        self.assertEqual(result, {Student[3], Student[4], Student[5]})
    def test_datetime_func3(self):
        result = set(select(s for s in Student if s.last_visit >= datetime(2011, 3, 3, 13, 13, 13)))
        self.assertEqual(result, {Student[3], Student[4], Student[5]})
    # @raises_exception(ExprEvalError, "datetime(1983, 'three', 3) raises TypeError: an integer is required")
    @raises_exception(TypeError, "'month' argument of datetime(...) function must be of 'int' type. "
                                 "Got: '%s'" % unicode.__name__)
    def test_datetime_func4(self):
        result = set(select(s for s in Student if s.last_visit >= datetime(1983, 'three', 3)))
    # @raises_exception(NotImplementedError)
    # def test_datetime_func5(self):
    #     d = 3
    #     result = set(select(s for s in Student if s.last_visit >= date(1983, d, 3)))
    def test_datetime_now1(self):
        result = set(select(s for s in Student if s.dob < date.today()))
        self.assertEqual(result, {Student[1], Student[2], Student[3], Student[4], Student[5]})
    @raises_exception(ExprEvalError, "`1 < datetime.now()` raises TypeError: " + (
        "can't compare 'datetime' to 'int'" if PYPY2 else
        "unorderable types: int < datetime" if PYPY else
        "can't compare datetime.datetime to int" if PY2 else
        "unorderable types: int() < datetime.datetime()" if sys.version_info < (3, 6) else
        "'<' not supported between instances of 'int' and 'datetime.datetime'"))
    def test_datetime_now2(self):
        select(s for s in Student if 1 < datetime.now())
    def test_datetime_now3(self):
        result = set(select(s for s in Student if s.dob < datetime.today()))
        self.assertEqual(result, {Student[1], Student[2], Student[3], Student[4], Student[5]})
    def test_decimal_func(self):
        result = set(select(s for s in Student if s.scholarship >= Decimal("303.3")))
        self.assertEqual(result, {Student[3], Student[4], Student[5]})
    def test_concat_1(self):
        result = set(select(concat(s.name, ':', s.dob.year, ':', s.scholarship) for s in Student))
        self.assertEqual(result, {'AA:1981:0', 'BB:1982:202.2', 'CC:1983:303.3', 'DD:1984:404.4', 'EE:1985:505.5'})
    @raises_exception(TranslationError, 'Invalid argument of concat() function: g.students')
    def test_concat_2(self):
        result = set(select(concat(g.number, g.students) for g in Group))

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function

import os, sys
from os.path import dirname

__version__ = '0.7.8'

def detect_mode():
    try: import google.appengine
    except ImportError: pass
    else:
        if os.environ.get('SERVER_SOFTWARE', '').startswith('Development'):
            return 'GAE-LOCAL'
        return 'GAE-SERVER'

    try: from mod_wsgi import version
    except: pass
    else: return 'MOD_WSGI'

    main = sys.modules['__main__']

    if not hasattr(main, '__file__'): # console
        return 'INTERACTIVE'

    if getattr(main, 'INTERACTIVE_MODE_AVAILABLE', False): # pycharm console
        return 'INTERACTIVE'

    if 'flup.server.fcgi' in sys.modules: return 'FCGI-FLUP'
    if 'uwsgi' in sys.modules: return 'UWSGI'
    if 'flask' in sys.modules: return 'FLASK'
    if 'cherrypy' in sys.modules: return 'CHERRYPY'
    if 'bottle' in sys.modules: return 'BOTTLE'
    return 'UNKNOWN'

MODE = detect_mode()

MAIN_FILE = None
if MODE == 'MOD_WSGI':
    for module_name, module in sys.modules.items():
        if module_name.startswith('_mod_wsgi_'):
            MAIN_FILE = module.__file__
            break
elif MODE != 'INTERACTIVE':
    MAIN_FILE = sys.modules['__main__'].__file__

if MAIN_FILE is not None: MAIN_DIR = dirname(MAIN_FILE)
else: MAIN_DIR = None

PONY_DIR = dirname(__file__)





class cached_property(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """  # noqa

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class class_property(object):
    """
    Read-only class property
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        return self.func(cls)

class class_cached_property(object):

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        value = self.func(cls)
        setattr(cls, self.func.__name__, value)
        return value


import unittest

from pony import orm

class Test(unittest.TestCase):
    def test_1(self):
        db = orm.Database('sqlite', ':memory:')

        class Person(db.Entity):
            name = orm.Required(str)

        db.generate_mapping(create_tables=True)

        with orm.db_session:
            a = Person(name='John')
            a.delete()
            Person.exists(name='Mike')



from __future__ import print_function

from setuptools import setup
import sys

import unittest

def test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('pony.orm.tests', pattern='test_*.py')
    return test_suite

name = "pony"
version = __import__('pony').__version__
description = "Pony Object-Relational Mapper"
long_description = """
About
=========
Pony ORM is easy to use and powerful object-relational mapper for Python.
Using Pony, developers can create and maintain database-oriented software applications
faster and with less effort. One of the most interesting features of Pony is
its ability to write queries to the database using generator expressions.
Pony then analyzes the abstract syntax tree of a generator and translates it
to its SQL equivalent.

Following is an example of a query in Pony::

    select(p for p in Product if p.name.startswith('A') and p.cost <= 1000)

Such approach simplify the code and allows a programmer to concentrate
on the business logic of the application.

Pony translates queries to SQL using a specific database dialect.
Currently Pony works with SQLite, MySQL, PostgreSQL and Oracle databases.

The package `pony.orm.examples <https://github.com/ponyorm/pony/tree/orm/pony/orm/examples>`_
contains several examples.

Installation
=================
::

    pip install pony

Entity-Relationship Diagram Editor
=============================================
`Pony online ER Diagram Editor <https://editor.ponyorm.com>`_ is a great tool for prototyping.
You can draw your ER diagram online, generate  Pony entity declarations or SQL script for
creating database schema based on the diagram and start working with the database in seconds.

Pony ORM Links:
=================
- Main site: https://ponyorm.com
- Documentation: https://docs.ponyorm.com
- GitHub: https://github.com/ponyorm/pony
- Mailing list:  http://ponyorm-list.ponyorm.com
- ER Diagram Editor: https://editor.ponyorm.com
- Blog: https://blog.ponyorm.com
"""

classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Topic :: Software Development :: Libraries',
    'Topic :: Database'
]

author = "Alexander Kozlovsky, Alexey Malashkevich"
author_email = "team@ponyorm.com"
url = "https://ponyorm.com"
licence = "Apache License Version 2.0"

packages = [
    "pony",
    "pony.flask",
    "pony.flask.example",
    "pony.orm",
    "pony.orm.dbproviders",
    "pony.orm.examples",
    "pony.orm.integration",
    "pony.orm.tests",
    "pony.thirdparty",
    "pony.thirdparty.compiler",
    "pony.utils"
]

package_data = {
    'pony.flask.example': ['templates/*.html'],
    'pony.orm.tests': ['queries.txt']
}

download_url = "http://pypi.python.org/pypi/pony/"

if __name__ == "__main__":
    pv = sys.version_info[:2]
    if pv not in ((2, 7), (3, 3), (3, 4), (3, 5), (3, 6), (3, 7)):
        s = "Sorry, but %s %s requires Python of one of the following versions: 2.7, 3.3-3.7." \
            " You have version %s"
        print(s % (name, version, sys.version.split(' ', 1)[0]))
        sys.exit(1)

    setup(
        name=name,
        version=version,
        description=description,
        long_description=long_description,
        classifiers=classifiers,
        author=author,
        author_email=author_email,
        url=url,
        license=licence,
        packages=packages,
        package_data=package_data,
        download_url=download_url,
        test_suite='setup.test_suite'
    )



from pony.py23compat import iteritems

import json
from datetime import date, datetime
from decimal import Decimal
from collections import defaultdict

from pony.orm.core import Entity, TransactionError
from pony.utils import cut_traceback, throw

class Bag(object):
    def __init__(bag, database):
        bag.database = database
        bag.session_cache = None
        bag.entity_configs = {}
        bag.objects = defaultdict(set)
        bag.vars = {}
        bag.dicts = defaultdict(dict)
    @cut_traceback
    def config(bag, entity, only=None, exclude=None, with_collections=True, with_lazy=False, related_objects=True):
        if bag.database.entities.get(entity.__name__) is not entity: throw(TypeError,
            'Entity %s does not belong to database %r' % (entity.__name__, bag.database))
        attrs = entity._get_attrs_(only, exclude, with_collections, with_lazy)
        bag.entity_configs[entity] = attrs, related_objects
        return attrs, related_objects
    @cut_traceback
    def put(bag, x):
        if isinstance(x, Entity):
            bag._put_object(x)
        else:
            try: x = list(x)
            except: throw(TypeError, 'Entity instance or a sequence of instances expected. Got: %r' % x)
            for item in x:
                if not isinstance(item, Entity): throw(TypeError,
                    'Entity instance or a sequence of instances expected. Got: %r' % item)
                bag._put_object(item)
    def _put_object(bag, obj):
        entity = obj.__class__
        if bag.database.entities.get(entity.__name__) is not entity: throw(TypeError,
            'Entity %s does not belong to database %r' % (entity.__name__, bag.database))
        cache = bag.session_cache
        if cache is None: cache = bag.session_cache = obj._session_cache_
        elif obj._session_cache_ is not cache: throw(TransactionError,
            'An attempt to mix objects belonging to different transactions')
        bag.objects[entity].add(obj)
    def _reduce_composite_pk(bag, pk):
        return ','.join(str(item).replace('*', '**').replace(',', '*,') for item in pk)
    @cut_traceback
    def to_dict(bag):
        bag.dicts.clear()
        for entity, objects in iteritems(bag.objects):
            for obj in objects:
                dicts = bag.dicts[entity]
                if obj not in dicts: bag._process_object(obj)
        result = defaultdict(dict)
        for entity, dicts in iteritems(bag.dicts):
            composite_pk = len(entity._pk_columns_) > 1
            for obj, d in iteritems(dicts):
                pk = obj._get_raw_pkval_()
                if composite_pk: pk = bag._reduce_composite_pk(pk)
                else: pk = pk[0]
                result[entity.__name__][pk] = d
        bag.dicts.clear()
        return result
    def _process_object(bag, obj, process_related=True):
        entity = obj.__class__
        try: attrs, related_objects = bag.entity_configs[entity]
        except KeyError: attrs, related_objects = bag.config(entity)
        process_related_objects = process_related and related_objects
        d = {}
        for attr in attrs:
            value = attr.__get__(obj)
            if attr.is_collection:
                if not process_related:
                    continue
                if process_related_objects:
                    for related_obj in value:
                        if related_obj not in bag.dicts:
                            bag._process_object(related_obj, process_related=False)
                if attr.reverse.entity._pk_is_composite_:
                    value = sorted(bag._reduce_composite_pk(item._get_raw_pkval_()) for item in value)
                else: value = sorted(item._get_raw_pkval_()[0] for item in value)
            elif attr.is_relation:
                if value is not None:
                    if process_related_objects:
                        bag._process_object(value, process_related=False)
                    value = value._get_raw_pkval_()
                    if len(value) == 1: value = value[0]
            d[attr.name] = value
        bag.dicts[entity][obj] = d
    @cut_traceback
    def to_json(bag):
        return json.dumps(bag.to_dict(), default=json_converter, indent=2, sort_keys=True)

def to_dict(objects):
    if isinstance(objects, Entity): objects = [ objects ]
    objects = iter(objects)
    try: first_object = next(objects)
    except StopIteration: return {}
    if not isinstance(first_object, Entity): throw(TypeError,
        'Entity instance or a sequence of instances expected. Got: %r' % first_object)
    database = first_object._database_
    bag = Bag(database)
    bag.put(first_object)
    bag.put(objects)
    return dict(bag.to_dict())

def to_json(objects):
    return json.dumps(to_dict(objects), default=json_converter, indent=2, sort_keys=True)

def json_converter(x):
    if isinstance(x, (datetime, date, Decimal)):
        return str(x)
    raise TypeError(x)



from __future__ import absolute_import, print_function, division

import unittest
from datetime import date, datetime

from pony.orm.core import *
from pony.orm.tests.testutils import *

db = Database('sqlite', ':memory:')

class Entity1(db.Entity):
	a = PrimaryKey(int)
	b = Required(date)
	c = Required(datetime)

db.generate_mapping(create_tables=True)

with db_session:
    Entity1(a=1, b=date(2009, 10, 20), c=datetime(2009, 10, 20, 10, 20, 30))
    Entity1(a=2, b=date(2010, 10, 21), c=datetime(2010, 10, 21, 10, 21, 31))
    Entity1(a=3, b=date(2011, 11, 22), c=datetime(2011, 11, 22, 10, 20, 32))

class TestDate(unittest.TestCase):
    def setUp(self):
        rollback()
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()
    def test_create(self):
        e1 = Entity1(a=4, b=date(2011, 10, 20), c=datetime(2009, 10, 20, 10, 20, 30))
    def test_date_year(self):
        result = select(e for e in Entity1 if e.b.year > 2009)
        self.assertEqual(len(result), 2)
    def test_date_month(self):
        result = select(e for e in Entity1 if e.b.month == 10)
        self.assertEqual(len(result), 2)
    def test_date_day(self):
        result = select(e for e in Entity1 if e.b.day == 22)
        self.assertEqual(len(result), 1)
    def test_datetime_year(self):
        result = select(e for e in Entity1 if e.c.year > 2009)
        self.assertEqual(len(result), 2)
    def test_datetime_month(self):
        result = select(e for e in Entity1 if e.c.month == 10)
        self.assertEqual(len(result), 2)
    def test_datetime_day(self):
        result = select(e for e in Entity1 if e.c.day == 22)
        self.assertEqual(len(result), 1)
    def test_datetime_hour(self):
        result = select(e for e in Entity1 if e.c.hour == 10)
        self.assertEqual(len(result), 3)
    def test_datetime_minute(self):
        result = select(e for e in Entity1 if e.c.minute == 20)
        self.assertEqual(len(result), 2)
    def test_datetime_second(self):
        result = select(e for e in Entity1 if e.c.second == 30)
        self.assertEqual(len(result), 1)

if __name__ == '__main__':
    unittest.main()




import unittest

from pony.orm import *
from pony import orm


class Test(unittest.TestCase):

    def setUp(self):
        db = self.db = Database('sqlite', ':memory:')

        class User(db.Entity):
            name = Required(str)
            servers = Set("Server")

        class Worker(db.User):
            pass

        class Admin(db.Worker):
            pass

        # And M:1 relationship with another entity
        class Server(db.Entity):
            name = Required(str)
            user = Optional(User)

        db.generate_mapping(check_tables=True, create_tables=True)

        with orm.db_session:
            Server(name='s1.example.com', user=User(name="Alex"))
            Server(name='s2.example.com', user=Worker(name="John"))
            Server(name='free.example.com', user=None)

    @db_session
    def test(self):
        qu = left_join((s.name, s.user.name) for s in self.db.Server)[:]
        for server, user in qu:
            if user is None:
                break
        else:
            self.fail()




from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.core import local
from pony.orm.tests.testutils import *

class TestGeneratorDbSession(unittest.TestCase):
    def setUp(self):
        db = Database('sqlite', ':memory:')
        class Account(db.Entity):
            id = PrimaryKey(int)
            amount = Required(int)
        db.generate_mapping(create_tables=True)

        self.db = db
        self.Account = Account

        with db_session:
            a1 = Account(id=1, amount=1000)
            a2 = Account(id=2, amount=2000)
            a3 = Account(id=3, amount=3000)

    def tearDown(self):
        assert local.db_session is None
        self.db = self.Account = None

    @raises_exception(TypeError, 'db_session with `retry` option cannot be applied to generator function')
    def test1(self):
        @db_session(retry=3)
        def f(): yield

    @raises_exception(TypeError, 'db_session with `ddl` option cannot be applied to generator function')
    def test2(self):
        @db_session(ddl=True)
        def f(): yield

    @raises_exception(TypeError, 'db_session with `serializable` option cannot be applied to generator function')
    def test3(self):
        @db_session(serializable=True)
        def f(): yield

    def test4(self):
        @db_session(immediate=True)
        def f(): yield

    @raises_exception(TransactionError, '@db_session-wrapped generator cannot be used inside another db_session')
    def test5(self):
        @db_session
        def f(): yield
        with db_session:
            next(f())

    def test6(self):
        @db_session
        def f():
            x = local.db_session
            self.assertTrue(x is not None)

            yield self.db._get_cache()
            self.assertEqual(local.db_session, x)

            a1 = self.Account[1]
            yield a1.amount
            self.assertEqual(local.db_session, x)

            a2 = self.Account[2]
            yield a2.amount

        gen = f()
        cache = next(gen)
        self.assertTrue(cache.is_alive)
        self.assertEqual(local.db_session, None)

        amount = next(gen)
        self.assertEqual(amount, 1000)
        self.assertEqual(local.db_session, None)

        amount = next(gen)
        self.assertEqual(amount, 2000)
        self.assertEqual(local.db_session, None)

        try: next(gen)
        except StopIteration:
            self.assertFalse(cache.is_alive)
        else:
            self.fail()

    def test7(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            id2 = yield a1.amount
            a2 = self.Account[id2]
            amount = yield a2.amount
            a1.amount -= amount
            a2.amount += amount
            commit()

        gen = f(1)

        amount1 = next(gen)
        self.assertEqual(amount1, 1000)

        amount2 = gen.send(2)
        self.assertEqual(amount2, 2000)

        try:
            gen.send(100)
        except StopIteration:
            pass
        else:
            self.fail()

        with db_session:
            a1 = self.Account[1]
            self.assertEqual(a1.amount, 900)
            a2 = self.Account[2]
            self.assertEqual(a2.amount, 2100)

    @raises_exception(TransactionError, 'You need to manually commit() changes before suspending the generator')
    def test8(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            a1.amount += 100
            yield a1.amount

        for amount in f(1):
            pass

    def test9(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            a1.amount += 100
            commit()
            yield a1.amount

        for amount in f(1):
            pass

    def test10(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            yield a1.amount
            a1.amount += 100

        with db_session:
            a = self.Account[1].amount
        for amount in f(1):
            pass
        with db_session:
            b = self.Account[1].amount

        self.assertEqual(b, a + 100)

    def test12(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            yield a1.amount

        gen = f(1)
        next(gen)
        gen.close()

    @raises_exception(TypeError, 'error message')
    def test13(self):
        @db_session
        def f(id1):
            a1 = self.Account[id1]
            yield a1.amount

        gen = f(1)
        next(gen)
        gen.throw(TypeError('error message'))

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import *

class TestInheritance(unittest.TestCase):

    def test_0(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            id = PrimaryKey(int)

        self.assertTrue(Entity1._root_ is Entity1)
        self.assertEqual(Entity1._discriminator_attr_, None)
        self.assertEqual(Entity1._discriminator_, None)

    def test_1(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            id = PrimaryKey(int)
        class Entity2(Entity1):
            a = Required(int)
        class Entity3(Entity1):
            b = Required(int)
        class Entity4(Entity2, Entity3):
            c = Required(int)

        self.assertTrue(Entity1._root_ is Entity1)
        self.assertTrue(Entity2._root_ is Entity1)
        self.assertTrue(Entity3._root_ is Entity1)
        self.assertTrue(Entity4._root_ is Entity1)
        self.assertTrue(Entity1._discriminator_attr_ is Entity1.classtype)
        self.assertTrue(Entity2._discriminator_attr_ is Entity1._discriminator_attr_)
        self.assertTrue(Entity3._discriminator_attr_ is Entity1._discriminator_attr_)
        self.assertTrue(Entity4._discriminator_attr_ is Entity1._discriminator_attr_)
        self.assertEqual(Entity1._discriminator_, 'Entity1')
        self.assertEqual(Entity2._discriminator_, 'Entity2')
        self.assertEqual(Entity3._discriminator_, 'Entity3')
        self.assertEqual(Entity4._discriminator_, 'Entity4')

    @raises_exception(ERDiagramError, "Multiple inheritance graph must be diamond-like. "
        "Entity Entity3 inherits from Entity1 and Entity2 entities which don't have common base class.")
    def test_2(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = PrimaryKey(int)
        class Entity2(db.Entity):
            b = PrimaryKey(int)
        class Entity3(Entity1, Entity2):
            c = Required(int)

    @raises_exception(ERDiagramError, 'Attribute Entity3.a conflicts with attribute Entity2.a '
                                      'because both entities inherit from Entity1. '
                                      'To fix this, move attribute definition to base class')
    def test_3a(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            id = PrimaryKey(int)
        class Entity2(Entity1):
            a = Required(int)
        class Entity3(Entity1):
            a = Required(int)

    def test3b(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            id = PrimaryKey(int)
        class Entity2(Entity1):
            a = Set('Entity4')
        class Entity3(Entity1):
            a = Set('Entity4')
        class Entity4(db.Entity):
            x = Required(Entity2)
            y = Required(Entity3)
        db.generate_mapping(create_tables=True)
        self.assertTrue(Entity2.a not in Entity1._subclass_attrs_)
        self.assertTrue(Entity3.a not in Entity1._subclass_attrs_)

    @raises_exception(ERDiagramError, "Name 'a' hides base attribute Entity1.a")
    def test_4(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            id = PrimaryKey(int)
            a = Required(int)
        class Entity2(Entity1):
            a = Required(int)

    @raises_exception(ERDiagramError, "Primary key cannot be redefined in derived classes")
    def test_5(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = PrimaryKey(int)
        class Entity2(Entity1):
            b = PrimaryKey(int)

    def test_6(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Discriminator(str)
            b = Required(int)
        class Entity2(Entity1):
            c = Required(int)

        self.assertTrue(Entity1._discriminator_attr_ is Entity1.a)
        self.assertTrue(Entity2._discriminator_attr_ is Entity1.a)

    @raises_exception(TypeError, "Discriminator value for entity Entity1 "
                                 "with custom discriminator column 'a' of 'int' type is not set")
    def test_7(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Discriminator(int)
            b = Required(int)
        class Entity2(Entity1):
            c = Required(int)

    def test_8(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            _discriminator_ = 1
            a = Discriminator(int)
            b = Required(int)
        class Entity2(Entity1):
            _discriminator_ = 2
            c = Required(int)
        db.generate_mapping(create_tables=True)
        with db_session:
            x = Entity1(b=10)
            y = Entity2(b=10, c=20)
        with db_session:
            x = Entity1[1]
            y = Entity1[2]
            self.assertTrue(isinstance(x, Entity1))
            self.assertTrue(isinstance(y, Entity2))
            self.assertEqual(x.a, 1)
            self.assertEqual(y.a, 2)

    def test_9(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            _discriminator_ = '1'
            a = Discriminator(int)
            b = Required(int)
        class Entity2(Entity1):
            _discriminator_ = '2'
            c = Required(int)
        db.generate_mapping(create_tables=True)
        with db_session:
            x = Entity1(b=10)
            y = Entity2(b=10, c=20)
        with db_session:
            x = Entity1[1]
            y = Entity1[2]
            self.assertTrue(isinstance(x, Entity1))
            self.assertTrue(isinstance(y, Entity2))
            self.assertEqual(x.a, 1)
            self.assertEqual(y.a, 2)

    @raises_exception(TypeError, "Incorrect discriminator value is set for Entity2 attribute 'a' of 'int' type: 'zzz'")
    def test_10(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            _discriminator_ = 1
            a = Discriminator(int)
            b = Required(int)
        class Entity2(Entity1):
            _discriminator_ = 'zzz'
            c = Required(int)

    def test_11(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            _discriminator_ = 1
            a = Discriminator(int)
            b = Required(int)
            composite_index(a, b)

    @raises_exception(ERDiagramError, 'Invalid use of attribute Entity1.a in entity Entity2')
    def test_12(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Required(int)
        class Entity2(db.Entity):
            b = Required(int)
            composite_index(Entity1.a, b)

    def test_13(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Required(int)
        class Entity2(Entity1):
            b = Required(int)
            composite_index(Entity1.a, b)
        self.assertEqual([ index.attrs for index in Entity2._indexes_ ],
                          [ (Entity2.id,), (Entity2.a, Entity2.b) ])

    def test_14(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            d = Discriminator(str)
            a = Required(int)
        class Entity2(Entity1):
            b = Required(int)
            composite_index(Entity1.d, Entity1.a, b)
        self.assertEqual([ index.attrs for index in Entity2._indexes_ ],
                          [ (Entity2.id,), (Entity2.d, Entity2.a, Entity2.b) ])

    def test_15(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            d = Discriminator(str)
            a = Required(int)
        class Entity2(Entity1):
            b = Required(int)
            composite_index('d', 'id', 'a', 'b')
        self.assertEqual([ index.attrs for index in Entity2._indexes_ ],
                         [ (Entity2.id,), (Entity2.d, Entity2.id, Entity2.a, Entity2.b) ])

    def test_16(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Required(int)
        class Entity2(Entity1):
            b = Required(int)
            composite_index('classtype', 'id', 'a', b)
        self.assertEqual([ index.attrs for index in Entity2._indexes_ ],
                         [ (Entity2.id,), (Entity2.classtype, Entity2.id, Entity2.a, Entity2.b) ])

    def test_17(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            t = Discriminator(str)
            a = Required(int)
            b = Required(int)
            composite_index(t, a, b)
        class Entity2(Entity1):
            c = Required(int)
        self.assertEqual([ index.attrs for index in Entity1._indexes_ ],
                         [ (Entity1.id,), (Entity1.t, Entity1.a, Entity1.b) ])

    def test_18(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Required(int)
        class Entity2(db.Entity1):
            b = Required(int)
        class Entity3(Entity1):
            c = Required(int)
        db.generate_mapping(create_tables=True)
        with db_session:
            x = Entity1(a=10)
            y = Entity2(a=20, b=30)
            z = Entity3(a=40, c=50)
        with db_session:
            result = select(e for e in Entity1 if e.b == 30 or e.c == 50)
            self.assertEqual([ e.id for e in result ], [ 2, 3 ])

    def test_discriminator_1(self):
        db = Database('sqlite', ':memory:')
        class Entity1(db.Entity):
            a = Discriminator(str)
            b = Required(int)
            PrimaryKey(a, b)
        class Entity2(db.Entity1):
            c = Required(int)
        db.generate_mapping(create_tables=True)
        with db_session:
            x = Entity1(b=10)
            y = Entity2(b=20, c=30)
        with db_session:
            obj = Entity1.get(b=20)
            self.assertEqual(obj.a, 'Entity2')
            self.assertEqual(obj.b, 20)
            self.assertEqual(obj._pkval_, ('Entity2', 20))
        with db_session:
            obj = Entity1['Entity2', 20]
            self.assertEqual(obj.a, 'Entity2')
            self.assertEqual(obj.b, 20)
            self.assertEqual(obj._pkval_, ('Entity2', 20))


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest
from decimal import Decimal
from datetime import date

from pony.orm.core import *
from pony.orm.tests.testutils import *

db = Database('sqlite', ':memory:')

class Group(db.Entity):
    id = PrimaryKey(int)
    major = Required(unicode)
    students = Set('Student')

class Student(db.Entity):
    name = Required(unicode)
    age = Optional(int)
    scholarship = Required(Decimal, default=0)
    picture = Optional(buffer, lazy=True)
    email = Required(unicode, unique=True)
    phone = Optional(unicode, unique=True)
    courses = Set('Course')
    group = Optional('Group')

class Course(db.Entity):
    name = Required(unicode)
    semester = Required(int)
    students = Set(Student)
    composite_key(name, semester)

db.generate_mapping(create_tables=True)

with db_session:
    g1 = Group(id=1, major='Math')
    g2 = Group(id=2, major='Physics')
    s1 = Student(id=1, name='S1', age=19, email='s1@example.com', group=g1)
    s2 = Student(id=2, name='S2', age=21, email='s2@example.com', group=g1)
    s3 = Student(id=3, name='S3', email='s3@example.com', group=g2)
    c1 = Course(name='Math', semester=1)
    c2 = Course(name='Math', semester=2)
    c3 = Course(name='Physics', semester=1)


class TestCRUD(unittest.TestCase):
    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_getitem_1(self):
        g1 = Group[1]
        self.assertEqual(g1.id, 1)

    @raises_exception(ObjectNotFound, 'Group[333]')
    def test_getitem_2(self):
        g333 = Group[333]

    def test_exists_1(self):
        x = Group.exists(id=1)
        self.assertEqual(x, True)

    def test_exists_2(self):
        x = Group.exists(id=333)
        self.assertEqual(x, False)

    def test_exists_3(self):
        g1 = Group[1]
        x = Student.exists(group=g1)
        self.assertEqual(x, True)

    def test_numeric_nonzero(self):
        result = select(s.id for s in Student if s.age)[:]
        self.assertEqual(set(result), {1, 2})

    def test_numeric_negate_1(self):
        result = select(s.id for s in Student if not s.age)[:]
        self.assertEqual(set(result), {3})
        self.assertTrue('is null' in db.last_sql.lower())

    def test_numeric_negate_2(self):
        result = select(c.id for c in Course if not c.semester)[:]
        self.assertEqual(result, [])
        self.assertTrue('is null' not in db.last_sql.lower())

    def test_set1(self):
        s1 = Student[1]
        s1.set(name='New name', scholarship=100)
        self.assertEqual(s1.name, 'New name')
        self.assertEqual(s1.scholarship, 100)

    def test_set2(self):
        g1 = Group[1]
        s3 = Student[3]
        g1.set(students=[s3])
        self.assertEqual(s3.group, Group[1])

    def test_set3(self):
        c1 = Course[1]
        c1.set(name='Algebra', semester=3)

    def test_set4(self):
        s1 = Student[1]
        s1.set(name='New name', email='new_email@example.com')

    def test_validate_1(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group=1)

    def test_validate_2(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group='1')

    @raises_exception(TransactionIntegrityError)
    def test_validate_3(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group=100)
        flush()

    @raises_exception(ValueError, "Value type for attribute Group.id must be int. Got string 'not a number'")
    def test_validate_5(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group='not a number')
    @raises_exception(TypeError, "Attribute Student.group must be of Group type. Got: datetime.date(2011, 1, 1)")
    def test_validate_6(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group=date(2011, 1, 1))

    @raises_exception(TypeError, 'Invalid number of columns were specified for attribute Student.group. Expected: 1, got: 2')
    def test_validate_7(self):
        s4 = Student(id=3, name='S4', email='s4@example.com', group=(1, 2))



import unittest
from decimal import Decimal
from datetime import date

from pony.orm import *
from pony.orm.serialization import to_dict
from pony.orm.tests.testutils import *

db = Database('sqlite', ':memory:')

class Student(db.Entity):
    name = Required(unicode)
    scholarship = Optional(int)
    gpa = Optional(Decimal, 3, 1)
    dob = Optional(date)
    group = Optional('Group')
    courses = Set('Course')
    biography = Optional(LongUnicode)

class Group(db.Entity):
    number = PrimaryKey(int)
    students = Set(Student)

class Course(db.Entity):
    name = Required(unicode, unique=True)
    students = Set(Student)

db.generate_mapping(create_tables=True)

with db_session:
    g1 = Group(number=1)
    g2 = Group(number=2)
    c1 = Course(name='Math')
    c2 = Course(name='Physics')
    c3 = Course(name='Computer Science')
    Student(id=1, name='S1', group=g1, gpa=3.1, courses=[c1, c2], biography='some text')
    Student(id=2, name='S2', group=g1, gpa=3.2, scholarship=100, dob=date(2000, 1, 1))
    Student(id=3, name='S3', group=g1, gpa=3.3, scholarship=200, dob=date(2001, 1, 2), courses=[c2, c3])
    Student(id=4, name='S4')

class TestObjectToDict(unittest.TestCase):
    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test1(self):
        s1 = Student[1]
        d = s1.to_dict()
        self.assertEqual(d, dict(id=1, name='S1', scholarship=None, gpa=Decimal('3.1'), dob=None,
                                 group=1))

    def test2(self):
        s1 = Student[1]
        d = s1.to_dict(related_objects=True)
        self.assertEqual(d, dict(id=1, name='S1', scholarship=None, gpa=Decimal('3.1'), dob=None,
                                 group=Group[1]))

    def test3(self):
        s1 = Student[1]
        d = s1.to_dict(with_collections=True)
        self.assertEqual(d, dict(id=1, name='S1', scholarship=None, gpa=Decimal('3.1'), dob=None,
                                 group=1, courses=[1, 2]))

    def test4(self):
        s1 = Student[1]
        d = s1.to_dict(with_collections=True, related_objects=True)
        self.assertEqual(d, dict(id=1, name='S1', scholarship=None, gpa=Decimal('3.1'), dob=None,
                                 group=Group[1], courses=[Course[1], Course[2]]))

    def test5(self):
        s1 = Student[1]
        d = s1.to_dict(with_lazy=True)
        self.assertEqual(d, dict(id=1, name='S1', scholarship=None, gpa=Decimal('3.1'), dob=None,
                                 group=1, biography='some text'))

    def test6(self):
        s1 = Student[1]
        d = s1.to_dict(only=['id', 'name', 'group'])
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test7(self):
        s1 = Student[1]
        d = s1.to_dict(['id', 'name', 'group'])
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test8(self):
        s1 = Student[1]
        d = s1.to_dict(only='id, name, group')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test9(self):
        s1 = Student[1]
        d = s1.to_dict(only='id name group')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test10(self):
        s1 = Student[1]
        d = s1.to_dict('id name group')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    @raises_exception(AttributeError, 'Entity Student does not have attriute x')
    def test11(self):
        s1 = Student[1]
        d = s1.to_dict('id name x group')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test12(self):
        s1 = Student[1]
        d = s1.to_dict('id name group', related_objects=True)
        self.assertEqual(d, dict(id=1, name='S1', group=Group[1]))

    def test13(self):
        s1 = Student[1]
        d = s1.to_dict(exclude=['dob', 'gpa', 'scholarship'])
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test14(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob, gpa, scholarship')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test15(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    @raises_exception(AttributeError, 'Entity Student does not have attriute x')
    def test16(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa x scholarship')
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test17(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship', related_objects=True)
        self.assertEqual(d, dict(id=1, name='S1', group=Group[1]))

    def test18(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship', with_lazy=True)
        self.assertEqual(d, dict(id=1, name='S1', group=1, biography='some text'))

    def test19(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship biography', with_lazy=True)
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test20(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship', with_collections=True)
        self.assertEqual(d, dict(id=1, name='S1', group=1, courses=[1, 2]))

    def test21(self):
        s1 = Student[1]
        d = s1.to_dict(exclude='dob gpa scholarship courses', with_collections=True)
        self.assertEqual(d, dict(id=1, name='S1', group=1))

    def test22(self):
        s1 = Student[1]
        d = s1.to_dict(only='id name group', exclude='dob group')
        self.assertEqual(d, dict(id=1, name='S1'))

    def test23(self):
        s1 = Student[1]
        d = s1.to_dict(only='id name group', exclude='dob group', with_collections=True, with_lazy=True)
        self.assertEqual(d, dict(id=1, name='S1'))

    def test24(self):
        c = Course(name='New Course')
        d = c.to_dict()  # should do flush and get c.id from the database
        self.assertEqual(d, dict(id=4, name='New Course'))

class TestSerializationToDict(unittest.TestCase):
    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test1(self):
        s4 = Student[4]
        self.assertEqual(s4.group, None)
        d = to_dict(s4)
        self.assertEqual(d, dict(Student={
            4 : dict(id=4, name='S4', group=None, dob=None, gpa=None, scholarship=None, courses=[])
            }))

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest, warnings
from datetime import date
from decimal import Decimal
from itertools import count

from pony.orm.core import *
from pony.orm.tests.testutils import *

class TestDBSession(unittest.TestCase):
    def setUp(self):
        self.db = Database('sqlite', ':memory:')
        class X(self.db.Entity):
            a = PrimaryKey(int)
            b = Optional(int)
        self.X = X
        self.db.generate_mapping(create_tables=True)
        with db_session:
            x1 = X(a=1, b=1)
            x2 = X(a=2, b=2)

    @raises_exception(TypeError, "Pass only keyword arguments to db_session or use db_session as decorator")
    def test_db_session_1(self):
        db_session(1, 2, 3)

    @raises_exception(TypeError, "Pass only keyword arguments to db_session or use db_session as decorator")
    def test_db_session_2(self):
        db_session(1, 2, 3, a=10, b=20)

    def test_db_session_3(self):
        self.assertTrue(db_session is db_session())

    def test_db_session_4(self):
        # Nested db_sessions are ignored
        with db_session:
            with db_session:
                self.X(a=3, b=3)
        with db_session:
            self.assertEqual(count(x for x in self.X), 3)

    def test_db_session_decorator_1(self):
        # Should commit changes on exit from db_session
        @db_session
        def test():
            self.X(a=3, b=3)
        test()
        with db_session:
            self.assertEqual(count(x for x in self.X), 3)

    def test_db_session_decorator_2(self):
        # Should rollback changes if an exception is occurred
        @db_session
        def test():
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_db_session_decorator_3(self):
        # Should rollback changes if the exception is not in the list of allowed exceptions
        @db_session(allowed_exceptions=[TypeError])
        def test():
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_db_session_decorator_4(self):
        # Should commit changes if the exception is in the list of allowed exceptions
        @db_session(allowed_exceptions=[ZeroDivisionError])
        def test():
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 3)
        else:
            self.fail()

    def test_allowed_exceptions_1(self):
        # allowed_exceptions may be callable, should commit if nonzero
        @db_session(allowed_exceptions=lambda e: isinstance(e, ZeroDivisionError))
        def test():
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 3)
        else:
            self.fail()

    def test_allowed_exceptions_2(self):
        # allowed_exceptions may be callable, should rollback if not nonzero
        @db_session(allowed_exceptions=lambda e: isinstance(e, TypeError))
        def test():
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    @raises_exception(TypeError, "'retry' parameter of db_session must be of integer type. Got: %r" % str)
    def test_retry_1(self):
        @db_session(retry='foobar')
        def test():
            pass

    @raises_exception(TypeError, "'retry' parameter of db_session must not be negative. Got: -1")
    def test_retry_2(self):
        @db_session(retry=-1)
        def test():
            pass

    def test_retry_3(self):
        # Should not to do retry until retry count is specified
        counter = count()
        @db_session(retry_exceptions=[ZeroDivisionError])
        def test():
            next(counter)
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            self.assertEqual(next(counter), 1)
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_retry_4(self):
        # Should rollback & retry 1 time if retry=1
        counter = count()
        @db_session(retry=1, retry_exceptions=[ZeroDivisionError])
        def test():
            next(counter)
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            self.assertEqual(next(counter), 2)
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_retry_5(self):
        # Should rollback & retry N time if retry=N
        counter = count()
        @db_session(retry=5, retry_exceptions=[ZeroDivisionError])
        def test():
            next(counter)
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            self.assertEqual(next(counter), 6)
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_retry_6(self):
        # Should not retry if the exception not in the list of retry_exceptions
        counter = count()
        @db_session(retry=3, retry_exceptions=[TypeError])
        def test():
            next(counter)
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            self.assertEqual(next(counter), 1)
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_retry_7(self):
        # Should commit after successful retrying
        counter = count()
        @db_session(retry=5, retry_exceptions=[ZeroDivisionError])
        def test():
            i = next(counter)
            self.X(a=3, b=3)
            if i < 2: 1/0
        try:
            test()
        except ZeroDivisionError:
            self.fail()
        else:
            self.assertEqual(next(counter), 3)
            with db_session:
                self.assertEqual(count(x for x in self.X), 3)

    @raises_exception(TypeError, "The same exception ZeroDivisionError cannot be specified "
                                 "in both allowed and retry exception lists simultaneously")
    def test_retry_8(self):
        @db_session(retry=3, retry_exceptions=[ZeroDivisionError],
                             allowed_exceptions=[ZeroDivisionError])
        def test():
            pass

    def test_retry_9(self):
        # retry_exceptions may be callable, should retry if nonzero
        counter = count()
        @db_session(retry=3, retry_exceptions=lambda e: isinstance(e, ZeroDivisionError))
        def test():
            i = next(counter)
            self.X(a=3, b=3)
            1/0
        try:
            test()
        except ZeroDivisionError:
            self.assertEqual(next(counter), 4)
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_retry_10(self):
        # Issue 313: retry on exception raised during db_session.__exit__
        retries = count()
        @db_session(retry=3)
        def test():
            next(retries)
            self.X(a=1, b=1)
        try:
            test()
        except TransactionIntegrityError:
            self.assertEqual(next(retries), 4)
        else:
            self.fail()

    @raises_exception(PonyRuntimeWarning, '@db_session decorator with `retry=3` option is ignored for test() function '
                                          'because it is called inside another db_session')
    def test_retry_11(self):
        @db_session(retry=3)
        def test():
            pass
        with warnings.catch_warnings():
            warnings.simplefilter('error', PonyRuntimeWarning)
            with db_session:
                test()

    def test_db_session_manager_1(self):
        with db_session:
            self.X(a=3, b=3)
        with db_session:
            self.assertEqual(count(x for x in self.X), 3)

    @raises_exception(TypeError, "@db_session can accept 'retry' parameter "
                      "only when used as decorator and not as context manager")
    def test_db_session_manager_2(self):
        with db_session(retry=3):
            self.X(a=3, b=3)

    def test_db_session_manager_3(self):
        # Should rollback if the exception is not in the list of allowed_exceptions
        try:
            with db_session(allowed_exceptions=[TypeError]):
                self.X(a=3, b=3)
                1/0
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 2)
        else:
            self.fail()

    def test_db_session_manager_4(self):
        # Should commit if the exception is in the list of allowed_exceptions
        try:
            with db_session(allowed_exceptions=[ZeroDivisionError]):
                self.X(a=3, b=3)
                1/0
        except ZeroDivisionError:
            with db_session:
                self.assertEqual(count(x for x in self.X), 3)
        else:
            self.fail()

    # restriction removed in 0.7.3:
    # @raises_exception(TypeError, "@db_session can accept 'ddl' parameter "
    #                   "only when used as decorator and not as context manager")
    def test_db_session_ddl_1(self):
        with db_session(ddl=True):
            pass

    def test_db_session_ddl_1a(self):
        with db_session(ddl=True):
              with db_session(ddl=True):
                  pass

    def test_db_session_ddl_1b(self):
        with db_session(ddl=True):
              with db_session:
                  pass

    @raises_exception(TransactionError, 'Cannot start ddl transaction inside non-ddl transaction')
    def test_db_session_ddl_1c(self):
        with db_session:
              with db_session(ddl=True):
                  pass

    @raises_exception(TransactionError, "@db_session-decorated test() function with `ddl` option "
                                        "cannot be called inside of another db_session")
    def test_db_session_ddl_2(self):
        @db_session(ddl=True)
        def test():
            pass
        with db_session:
            test()

    def test_db_session_ddl_3(self):
        @db_session(ddl=True)
        def test():
            pass
        test()

    @raises_exception(ZeroDivisionError)
    def test_db_session_exceptions_1(self):
        def before_insert(self):
            1/0
        self.X.before_insert = before_insert
        with db_session:
            self.X(a=3, b=3)
            # Should raise ZeroDivisionError and not CommitException

    @raises_exception(ZeroDivisionError)
    def test_db_session_exceptions_2(self):
        def before_insert(self):
            1 / 0
        self.X.before_insert = before_insert
        with db_session:
            self.X(a=3, b=3)
            commit()
            # Should raise ZeroDivisionError and not CommitException

    @raises_exception(ZeroDivisionError)
    def test_db_session_exceptions_3(self):
        def before_insert(self):
            1 / 0
        self.X.before_insert = before_insert
        with db_session:
            self.X(a=3, b=3)
            db.commit()
            # Should raise ZeroDivisionError and not CommitException

    @raises_exception(ZeroDivisionError)
    def test_db_session_exceptions_4(self):
        with db_session:
            connection = self.db.get_connection()
            connection.close()
            1/0

db = Database('sqlite', ':memory:')

class Group(db.Entity):
    id = PrimaryKey(int)
    major = Required(unicode)
    students = Set('Student')

class Student(db.Entity):
    name = Required(unicode)
    picture = Optional(buffer, lazy=True)
    group = Required('Group')

db.generate_mapping(create_tables=True)

with db_session:
    g1 = Group(id=1, major='Math')
    g2 = Group(id=2, major='Physics')
    s1 = Student(id=1, name='S1', group=g1)
    s2 = Student(id=2, name='S2', group=g1)
    s3 = Student(id=3, name='S3', group=g2)


class TestDBSessionScope(unittest.TestCase):
    def setUp(self):
        rollback()

    def tearDown(self):
        rollback()

    def test1(self):
        with db_session:
            s1 = Student[1]
        name = s1.name

    @raises_exception(DatabaseSessionIsOver, 'Cannot load attribute Student[1].picture: the database session is over')
    def test2(self):
        with db_session:
            s1 = Student[1]
        picture = s1.picture

    @raises_exception(DatabaseSessionIsOver, 'Cannot load attribute Group[1].major: the database session is over')
    def test3(self):
        with db_session:
            s1 = Student[1]
        group_id = s1.group.id
        major = s1.group.major

    @raises_exception(DatabaseSessionIsOver, 'Cannot assign new value to Student[1].name: the database session is over')
    def test4(self):
        with db_session:
            s1 = Student[1]
        s1.name = 'New name'

    def test5(self):
        with db_session:
            g1 = Group[1]
        self.assertEqual(str(g1.students), 'StudentSet([...])')

    @raises_exception(DatabaseSessionIsOver, 'Cannot load collection Group[1].students: the database session is over')
    def test6(self):
        with db_session:
            g1 = Group[1]
        l = len(g1.students)

    @raises_exception(DatabaseSessionIsOver, 'Cannot change collection Group[1].students: the database session is over')
    def test7(self):
        with db_session:
            s1 = Student[1]
            g1 = Group[1]
        g1.students.remove(s1)

    @raises_exception(DatabaseSessionIsOver, 'Cannot change collection Group[1].students: the database session is over')
    def test8(self):
        with db_session:
            g2_students = Group[2].students
            g1 = Group[1]
        g1.students = g2_students

    @raises_exception(DatabaseSessionIsOver, 'Cannot change collection Group[1].students: the database session is over')
    def test9(self):
        with db_session:
            s3 = Student[3]
            g1 = Group[1]
        g1.students.add(s3)

    @raises_exception(DatabaseSessionIsOver, 'Cannot change collection Group[1].students: the database session is over')
    def test10(self):
        with db_session:
            g1 = Group[1]
        g1.students.clear()

    @raises_exception(DatabaseSessionIsOver, 'Cannot delete object Student[1]: the database session is over')
    def test11(self):
        with db_session:
            s1 = Student[1]
        s1.delete()

    @raises_exception(DatabaseSessionIsOver, 'Cannot change object Student[1]: the database session is over')
    def test12(self):
        with db_session:
            s1 = Student[1]
        s1.set(name='New name')

    def test_db_session_strict_1(self):
        with db_session(strict=True):
            s1 = Student[1]

    @raises_exception(DatabaseSessionIsOver, 'Cannot read value of Student[1].name: the database session is over')
    def test_db_session_strict_2(self):
        with db_session(strict=True):
            s1 = Student[1]
        name = s1.name

if __name__ == '__main__':
    unittest.main()
